from z3 import Solver, sat

from flamapy.core.operations import Operation
from flamapy.metamodels.smt_metamodel.models import PySMTModel


class ValidModel(Operation):

    def __init__(self, file_name: str) -> None:
        self.file_name: str = file_name
        self.result: bool = True

    def get_result(self) -> bool:
        return self.result

    def execute(self, model: PySMTModel) -> None:
        formula = model.domains[self.file_name]
        solver = Solver()
        solver.add(formula)
        self.result = solver.check() == sat


from z3 import BoolRef, ArithRef

from flamapy.core.models import VariabilityModel


class PySMTModel(VariabilityModel):

    @staticmethod
    def get_extension() -> str:
        return 'pysmt'

    def __init__(self) -> None:
        self.domains: dict[str, BoolRef] = {}
        self.cvvs: dict[str, ArithRef] = {}


from z3 import ModelRef, RatNumRef, IntNumRef


def config_sanitizer(config: ModelRef) -> dict[str, float | int]:
    sanitize_config: dict[str, float | int] = {}
    for var in config:
        value = 0
        if 'CVE' in str(var) or '/0' in str(var):
            continue
        if isinstance(config[var], RatNumRef):
            value = config[var].numerator_as_long() / config[var].denominator_as_long()
        elif isinstance(config[var], IntNumRef):
            value = config[var].as_long()
        sanitize_config[str(var)] = value
    return sanitize_config


from z3 import Int, sat, Optimize

from flamapy.core.operations import Operation
from flamapy.metamodels.smt_metamodel.models import PySMTModel
from flamapy.metamodels.smt_metamodel.utils import config_sanitizer


class CompleteConfig(Operation):

    def __init__(
        self,
        file_name: str,
        config: dict[str, int]
    ) -> None:
        self.file_name: str = file_name
        self.config: dict[str, int] = config
        self.result: dict[str, float | int] = {}

    def get_result(self) -> dict[str, float | int]:
        return self.result

    def execute(self, model: PySMTModel) -> None:
        solver = Optimize()
        if model.cvvs:
            cvss_f = model.cvvs[self.file_name]
            solver.minimize(cvss_f)

        formula = model.domains[self.file_name]
        solver.add(formula)
        for package, count in self.config.items():
            solver.add(Int(package) == count)

        while solver.check() == sat:
            config = solver.model()
            sanitized_config = config_sanitizer(config)
            self.result = sanitized_config
            break


from z3 import Or, Solver, sat

from flamapy.core.operations import Operation
from flamapy.metamodels.smt_metamodel.models import PySMTModel


class NumberOfProducts(Operation):

    def __init__(self, file_name: str) -> None:
        self.file_name: str = file_name
        self.result: int = 0

    def get_result(self) -> int:
        return self.result

    def execute(self, model: PySMTModel) -> None:
        formula = model.domains[self.file_name]
        solver = Solver()
        solver.add(formula)
        while solver.check() == sat:
            config = solver.model()

            block = []
            for var in config:
                if str(var) != '/0':
                    variable = var()
                    if 'CVSS' not in str(variable):
                        block.append(config[var] != variable)

            solver.add(Or(block))
            self.result += 1


from .valid_model import ValidModel
from .number_of_products import NumberOfProducts
from .filter_configs import FilterConfigs
from .maximize_impact import MaximizeImpact
from .minimize_impact import MinimizeImpact
from .valid_config import ValidConfig
from .complete_config import CompleteConfig
from .config_by_impact import ConfigByImpact

__all__ = [
    'ValidModel',
    'NumberOfProducts',
    'FilterConfigs',
    'MaximizeImpact',
    'MinimizeImpact',
    'ValidConfig',
    'CompleteConfig',
    'ConfigByImpact'
]


from z3 import And, ArithRef, BoolRef, Implies, Int, Or, Real, Not

from flamapy.core.transformations import ModelToModel
from flamapy.metamodels.dn_metamodel.models import (
    DependencyNetwork,
    Package,
    Version
)
from flamapy.metamodels.smt_metamodel.models import PySMTModel


class NetworkToSMT(ModelToModel):

    @staticmethod
    def get_source_extension() -> str:
        return 'dn'

    @staticmethod
    def get_destination_extension() -> str:
        return 'smt'

    def __init__(self, source_model: DependencyNetwork, agregator: str | None = None) -> None:
        self.source_model: DependencyNetwork = source_model
        self.agregator: str | None = agregator
        self.destination_model: PySMTModel = PySMTModel()
        self.vars: dict[str, ArithRef] = {}
        self.childs: dict[BoolRef, list[BoolRef]] = {}
        self.parents: dict[ArithRef, list[BoolRef]] = {}
        self.cvss_p: list[ArithRef] = []
        self.domain: list[BoolRef] = []
        self.distributions: list[str] = []
        self.ctcs: list[BoolRef] = []
        self.directs: list[str] = []

    def transform(self) -> None:
        if self.source_model.requirement_files:
            for requirement_file in self.source_model.requirement_files:
                self.transform_direct_packages(requirement_file.packages)
                self.build_indirect_constraints()
                self.domain.extend(self.ctcs)

                cvss_f_name = 'CVSS' + requirement_file.name
                cvss_f_var = Real(cvss_f_name)
                agregator_impact = self.agregate(self.cvss_p) if self.cvss_p else 0.
                self.domain.append(cvss_f_var == agregator_impact)

                func_obj_name = 'func_obj_' + requirement_file.name
                func_obj_var = Real(func_obj_name)
                func_obj_impact = self.obj_func(self.cvss_p) if self.cvss_p else 0.
                self.domain.append(func_obj_var == func_obj_impact)
                self.destination_model.cvvs[requirement_file.name] = func_obj_var

                self.destination_model.domains[requirement_file.name] = And(self.domain)
                self.domain.clear()
                self.cvss_p.clear()

    def transform_direct_packages(self, packages: list[Package]) -> None:
        for package in packages:
            self.directs.append(package.name)
        for package in packages:
            if package.name not in self.vars:
                var = Int(package.name)
                self.vars[package.name] = var

                cvss_p_name = 'CVSS' + package.name
                cvss_p_var = Real(cvss_p_name)
                self.vars[cvss_p_name] = cvss_p_var
                self.cvss_p.append(cvss_p_var)
            else:
                var = self.vars[package.name]
                cvss_p_var = self.vars['CVSS' + package.name]

            self.build_direct_contraint(var, package.versions)

            self.transform_versions(package.versions, var, cvss_p_var)

    def transform_versions(
        self,
        versions: list[Version],
        var: ArithRef,
        cvss_p_var: ArithRef
    ) -> None:
        for version in versions:
            key = str(var) + '-' + str(version.count)
            if key not in self.distributions:
                impacts = self.get_impacts(version)
                v_impact = self.agregate(impacts) if impacts else 0.
                ctc = Implies(var == version.count, cvss_p_var == v_impact)
                self.ctcs.append(ctc)
                self.distributions.append(key)
                self.transform_indirect_packages(version.packages, var, version.count)

    def transform_indirect_packages(
        self,
        packages: list[Package],
        parent: ArithRef,
        version: int
    ) -> None:
        for package in packages:
            if package.name not in self.vars:
                var = Int(package.name)
                self.vars[package.name] = var

                cvss_p_name = 'CVSS' + package.name
                cvss_p_var = Real(cvss_p_name)
                self.vars[cvss_p_name] = cvss_p_var
                self.cvss_p.append(cvss_p_var)

                ctc = Implies(var == -1, cvss_p_var == 0.)
                self.ctcs.append(ctc)
            else:
                var = self.vars[package.name]
                cvss_p_var = self.vars['CVSS' + package.name]

            self.append_indirect_constraint(var, package.versions, parent, version)

            self.transform_versions(package.versions, var, cvss_p_var)

    def get_impacts(self, version: Version) -> list[float]:
        impacts: list[float] = []

        for cve in version.cves:
            for key, value in cve['metrics'].items():
                match key:
                    case 'cvssMetricV31':
                        impacts.append(float(value[0]['impactScore']))
                    case 'cvssMetricV30':
                        impacts.append(float(value[0]['impactScore']))
                    case 'cvssMetricV2':
                        impacts.append(float(value[0]['impactScore']))

        return impacts

    def build_direct_contraint(self, var: ArithRef, versions: list[Version]) -> None:
        constraint = [var == version.count for version in versions]
        if constraint:
            self.domain.append(Or(constraint))

    def append_indirect_constraint(
        self,
        var: ArithRef,
        versions: list[Version],
        parent: ArithRef,
        version: int
    ) -> None:
        _ = versions[0]
        parts = [var == version.count for version in versions]
        if parts:
            versions = Or(parts)

            if versions in self.childs:
                self.childs[versions].append(parent == version)
            else:
                self.childs[versions] = [parent == version]

        if str(var) not in self.directs:
            if var in self.parents:
                self.parents[var].append(parent == version)
            else:
                self.parents[var] = [parent == version]

    def build_indirect_constraints(self) -> None:
        for versions, parents in self.childs.items():
            self.domain.append(Implies(Or(parents), versions))
        for var, parents in self.parents.items():
            self.domain.append(Implies(Not(Or(parents)), var == -1))

    # TODO: Posibilidad de añadir nuevas métricas
    def agregate(
        self,
        impacts: list[ArithRef | float],

    ) -> float:
        match self.agregator:
            case 'mean':
                return self.mean(impacts)
            case 'weighted_mean':
                return self.weighted_mean(impacts)
            case _:
                return self.mean(impacts)

    @staticmethod
    def obj_func(problems: list[ArithRef | float]) -> float:
        return sum(problems)

    @staticmethod
    def mean(problems: list[ArithRef | float]) -> float:
        return sum(problems) / len(problems)

    @staticmethod
    def weighted_mean(problems: list[ArithRef | float]) -> float:
        dividends = 0.
        divisors = 0.

        for var in problems:
            weight = var * 0.1
            dividends += var * weight
            divisors += weight

        return dividends / divisors


import setuptools

with open('README.md', 'r') as fh:
    long_description = fh.read()

setuptools.setup(
    name='flamapy-smt',
    version='1.2.2',
    author='Antonio Germán Márquez Trujillo',
    author_email='amtrujillo@us.es',
    description='This repo host the SMT model concrete classes',
    long_description=long_description,
    long_description_content_type='text/markdown',
    url='https://github.com/flamapy/smt_metamodel',
    packages=setuptools.find_namespace_packages(include=['flamapy.*']),
    classifiers=[
        'Programming Language :: Python :: 3',
        'License :: OSI Approved :: GNU General Public License v3 or later (GPLv3+)',
        'Operating System :: OS Independent'
    ],
    python_requires='>=3.10',
    install_requires=[
        'z3-solver==4.11.2.0',
        'flamapy==1.0.1',
        'flamapy-dn==1.1.6'
    ],
    tests_requires=[
        'prospector[with_everything]==1.8.3',
        'mypy==0.982',
        'types-setuptools==65.4.0'
    ]
)



from z3 import Or, Optimize, sat

from flamapy.core.operations import Operation
from flamapy.metamodels.smt_metamodel.models import PySMTModel
from flamapy.metamodels.smt_metamodel.utils import config_sanitizer


class MaximizeImpact(Operation):

    def __init__(
        self,
        file_name: str,
        limit: int
    ) -> None:
        self.file_name: str = file_name
        self.limit: int = limit
        self.result: list[dict[str, float | int]] = []

    def get_result(self) -> list[dict[str, float | int]]:
        return self.result

    def execute(self, model: PySMTModel) -> None:
        solver = Optimize()
        if model.cvvs:
            cvss_f = model.cvvs[self.file_name]
            solver.maximize(cvss_f)

        formula = model.domains[self.file_name]
        solver.add(formula)
        while len(self.result) < self.limit and solver.check() == sat:
            config = solver.model()
            sanitized_config = config_sanitizer(config)
            self.result.append(sanitized_config)

            block = []
            for var in config:
                if str(var) != '/0':
                    variable = var()
                    if 'CVSS' not in str(variable):
                        block.append(config[var] != variable)

            solver.add(Or(block))